# -*- coding: utf-8 -*-
"""
Useful global utilities.
"""

import warnings


def apply_default_permissions(instance):
    from .models import Object
    from .models.verb import AccessibleVerb

    set_default_permissions = AccessibleVerb.objects.filter(
        origin=Object.objects.get(pk=1), names__name="set_default_permissions"
    )
    if set_default_permissions:
        set_default_permissions[0](instance)
    else:
        warnings.warn(f"set_default_permissions failed for {instance}: verb not found", category=RuntimeWarning)



#!moo verb go --on "player class" --ability

from moo.core import api

direction = api.parser.get_dobj_str()

if api.caller.location.has_property("exits"):
    exits = api.caller.location.get_property("exits")
else:
    exits = {}

if direction not in exits:
    print("[red]There is no exit in that direction.[/red]")
    return  # pylint: disable=return-outside-function  # type: ignore

exit_info = exits[direction]
destination = exit_info["destination"]
door = exit_info.get("door")

if door and door.is_locked():
    print(f"[red]The {door.name} is locked.[/red]")
    return  # pylint: disable=return-outside-function  # type: ignore

api.caller.location = destination
api.caller.save()
print(f"You go {direction}.")



#!moo verb make --on "author class" --ability

from moo.core import api, create, lookup

if not (api.parser.has_dobj_str()):
    print("[color yellow]What do you want to make?[/color yellow]")
    return  # pylint: disable=return-outside-function  # type: ignore

name = api.parser.get_dobj_str()
new_obj = create(name)
print("[color yellow]Created %s[/color yellow]" % new_obj)

if api.parser.has_pobj_str("from"):
    parent_name = api.parser.get_pobj_str("from")
    try:
        new_obj.parent = lookup(f"{parent_name} class")
    except new_obj.DoesNotExist:
        print(f"[color red]No such object: {parent_name}[/color red]")
        return  # pylint: disable=return-outside-function  # type: ignore
    new_obj.save()
    print("[color yellow]Transmuted %s to %s[/color yellow]" % (new_obj, new_obj.parent))



#!moo verb describe --on "container class" --method

from moo.core import lookup

obj = args[1]  # pylint: disable=undefined-variable  # type: ignore
system = lookup("system object")

response = f"[bright_yellow]{obj.name}[/bright_yellow]\n"
response += system.describe(obj)

contents = obj.contents.filter(obvious=True)
if contents:
    response += "\n[yellow]Obvious contents:[/yellow]\n"
    for content in contents:
        response += f"{content.name}\n"

return response  # pylint: disable=return-outside-function  # type: ignore



#!moo verb test-async-verbs --on "player class" --ability --method

from moo.core import api, invoke

counter = 1
if args and len(args):  # pylint: disable=undefined-variable
    counter = args[0] + 1  # pylint: disable=undefined-variable

print(counter)

if counter < 10:
    verb = api.caller.get_verb("test-async-verbs")
    invoke(counter, verb=verb)



import pytest

from .. import code, parse
from ..models import Object, Verb


@pytest.mark.django_db
def test_caller_can_invoke_trivial_verb(t_init: Object, t_wizard: Object):
    printed = []
    description = t_wizard.location.properties.get(name="description")

    def _writer(msg):
        printed.append(msg)

    with code.context(t_wizard, _writer):
        writer = code.context.get("writer")
        globals = code.get_default_globals()  # pylint: disable=redefined-builtin
        globals.update(code.get_restricted_environment(writer))
        src = "from moo.core import api\napi.caller.invoke_verb('inspect')"
        code.r_exec(src, {}, globals)
        assert printed == [description.value]


@pytest.mark.django_db
def test_args_is_null_when_using_parser(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.context(t_wizard, _writer):
        parse.interpret("test-args-parser")
    assert printed == ["PARSER"]


@pytest.mark.django_db
def test_args_is_not_null_when_using_eval(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    verb = Verb.objects.get(names__name="test-args")
    with code.context(t_wizard, _writer):
        code.interpret(verb.code)
    assert printed == ["METHOD:():{}"]


@pytest.mark.django_db
def test_args_when_calling_multiple_verbs(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.context(t_wizard, _writer):
        parse.interpret("test-nested-verbs")
    assert printed == list(range(1, 11))


@pytest.mark.django_db
def test_write_to_caller(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.context(t_wizard, _writer):
        with pytest.warns(RuntimeWarning, match=r"ConnectionError\(\#3 \(Wizard\)\)\: TEST_STRING"):
            code.interpret("from moo.core import api, write\nwrite(api.caller, 'TEST_STRING')")



import pytest

from moo.core import code, lookup, parse
from moo.core.models import Object


@pytest.mark.django_db
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_creation(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.context(t_wizard, _writer):
        parse.interpret("make a widget")
        widget = lookup("widget")
        assert widget.location == t_wizard.location
        assert printed == [
            f"[color yellow]Created #{widget.id} (widget)[/color yellow]",
        ]


@pytest.mark.django_db
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_transmutation(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.context(t_wizard, _writer):
        parse.interpret("make a jar from container")
        jar = lookup("jar")
        container = lookup("container class")
        assert printed == [
            f"[color yellow]Created #{jar.id} (jar)[/color yellow]",
            f"[color yellow]Transmuted #{jar.id} (jar) to #{container.id} (container class)[/color yellow]",
        ]


@pytest.mark.django_db
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_description(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.context(t_wizard, _writer):
        parse.interpret("make a thingy")
        thingy = lookup("thingy")
        assert printed == [
            f"[color yellow]Created #{thingy.id} (thingy)[/color yellow]",
        ]
        printed.clear()

        parse.interpret("describe")
        parse.interpret("describe thingy")
        assert printed == [
            "[red]What do you want to describe?[/red]",
            "[red]What do you want to describe that as?[/red]",
        ]
        printed.clear()

        parse.interpret("describe thingy as 'a dusty old widget'")
        parse.interpret("look at thingy")
        print(printed)
        assert printed == [
            f"[color yellow]Description set for #{thingy.id} (thingy)[/color yellow]",
            "[bright_yellow]thingy[/bright_yellow]\n[deep_sky_blue1]a dusty old widget[/deep_sky_blue1]",
        ]



import importlib.resources

from django.conf import settings
from django.core.management.base import BaseCommand

from moo.core.bootstrap import load_python
from moo.core.models import Repository

builtin_templates = ["minimal", "default"]


class Command(BaseCommand):
    help = "Initialize a new database."

    def add_arguments(self, parser):
        parser.add_argument(
            "--bootstrap",
            type=str,
            default="default",
            help="Optionally pass a built-in template name or a Python source file" " to bootstrap the database.",
        )

    def handle(self, *args, bootstrap="default", **config):
        try:
            Repository.objects.get(slug=bootstrap)
            raise RuntimeError("Looks like this database has already been initialized.")
        except Repository.DoesNotExist:
            Repository.objects.create(
                slug=bootstrap, prefix=f"moo/core/bootstrap/{bootstrap}_verbs", url=settings.DEFAULT_GIT_REPO_URL
            )
        if bootstrap in builtin_templates:
            ref = importlib.resources.files("moo.core.bootstrap") / f"{bootstrap}.py"
            with importlib.resources.as_file(ref) as path:
                bootstrap_path = path
        else:
            raise NotImplementedError(bootstrap)
        load_python(bootstrap_path)



# -*- coding: utf-8 -*-
"""
Support utilities for database boostrapping.
"""

import argparse
import importlib.resources
import logging
import shlex
import warnings

from django.conf import settings

log = logging.getLogger(__name__)

parser = argparse.ArgumentParser("moo")
parser.add_argument("subcommand", choices=["verb"])
parser.add_argument("names", nargs="+")
parser.add_argument("--on", help="The object to add or modify the verb on")
parser.add_argument("--ability", action="store_true", help="Whether the verb is an intrinsic ability")
parser.add_argument("--method", action="store_true", help="Whether the verb is a method (callable from verb code)")


def get_source(filename, dataset="default"):
    """
    Get the source code for a verb from a Python package.

    :param filename: The name of the file to get the source code for.
    :type filename: str
    :param dataset: The name of the dataset to get the source code for.
    :type dataset: str
    :return: The source code for the verb.
    :rtype: str
    """
    ref = importlib.resources.files("moo.core.bootstrap") / f"{dataset}_verbs/{filename}"
    with importlib.resources.as_file(ref) as path:
        with open(path, encoding="utf8") as f:
            return f.read()


def load_python(python_path):
    """
    Execute a provided Python bootstrap file against the provided database.

    :param python_path: The path to the Python file to execute.
    :type python_path: str
    """
    with open(python_path, encoding="utf8") as f:
        src = f.read()
        exec(compile(src, python_path, "exec"), globals(), dict())  # pylint: disable=exec-used


def initialize_dataset(dataset="default"):
    """
    Initialize a new dataset.

    This will create the default objects and permissions for the dataset. Notably, it will
    create a `System Object` that is used to store global properties and verbs.

    It will also create a `Wizard` user that is used to manage the system.

    :param dataset: The name of the dataset to initialize.
    :type dataset: str
    :return: The repository object for the dataset.
    :rtype: Repository
    """
    from moo.core import create, models

    for name in settings.DEFAULT_PERMISSIONS:
        _ = models.Permission.objects.create(name=name)
    repo = models.Repository.objects.get(slug=dataset)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        system = create(name="System Object", unique_name=True)
        set_default_permissions = models.Verb.objects.create(
            method=True,
            origin=system,
            repo=repo,
            code=get_source("_system_set_default_permissions.py", dataset=dataset),
        )
        set_default_permissions.names.add(
            models.VerbName.objects.create(verb=set_default_permissions, name="set_default_permissions")
        )
        set_default_permissions(set_default_permissions)
        set_default_permissions(system)
    containers = create(name="container class", unique_name=True)
    containers.add_verb("accept", code="return True", method=True)
    # Create the first real user
    wizard = create(name="Wizard", unique_name=True, parents=[containers])
    wizard.owner = wizard
    wizard.save()
    # Wizard owns containers
    containers.owner = wizard
    containers.save()
    # Wizard owns the system...
    system.owner = wizard
    system.save()
    # ...and the default permissions verb
    set_default_permissions.owner = wizard
    set_default_permissions.save()
    return repo


def load_verbs(repo, verb_package):
    """
    Load the verbs from a Python package into the database and associate them with the given repository.

    Verb files should start with a shebang:

    .. code-block::

            #!moo [-h] [--on ON] [--ability] [--method] {verb} names [names ...]

            positional arguments:
            {verb}
            names

            options:
            -h, --help  show this help message and exit
            --on ON     The object to add or modify the verb on
            --ability   Whether the verb is an intrinsic ability
            --method    Whether the verb is a method (callable from verb code)

    :param repo: The repository object for the dataset.
    :type repo: Repository
    :param verb_package: The Python package to load the verbs from.
    :type verb_package: str
    """
    from moo.core.models.object import Object

    for ref in importlib.resources.files(verb_package).iterdir():
        if not ref.is_file():
            continue

        with importlib.resources.as_file(ref) as path:
            with open(path, encoding="utf8") as f:
                contents = f.read()
                try:
                    first, _ = contents.split("\n", maxsplit=1)
                except ValueError:
                    continue
                if first.startswith("#!moo "):
                    log.info(f"Loading verb source `{ref.name}`...")
                    args = parser.parse_args(shlex.split(first[6:]))
                    obj = Object.objects.get(name=args.on)
                    obj.add_verb(
                        *args.names,
                        code=contents,
                        filename=str(path.resolve()),
                        repo=repo,
                        ability=args.ability,
                        method=args.method,
                    )
                else:
                    log.info(f"Skipping verb source `{ref.name}`...")



#!moo verb describe --on "author class" --ability

from moo.core import api

if not (api.parser.has_dobj_str()):
    print("[red]What do you want to describe?[/red]")
    return  # pylint: disable=return-outside-function  # type: ignore
if not (api.parser.has_pobj_str("as")):
    print("[red]What do you want to describe that as?[/red]")
    return  # pylint: disable=return-outside-function  # type: ignore

subject = api.parser.get_dobj()
subject.set_property("description", api.parser.get_pobj_str("as"))
print("[color yellow]Description set for %s[/color yellow]" % subject)



#!moo verb dig --on "author class" --ability

from moo.core import api, create, lookup

directions = ["north", "northeast", "east", "southeast", "south", "southwest", "west", "northwest", "up", "down"]

if api.parser.has_pobj("through"):
    door = api.parser.get_pobj("through")
else:
    door = None

if api.caller.location.has_property("exits"):
    exits = api.caller.location.get_property("exits")
else:
    exits = {}

direction = api.parser.get_dobj_str()

if direction in exits:
    print("[color red]There is already an exit in that direction.[/color red]")
    return  # pylint: disable=return-outside-function  # type: ignore

room_class = lookup("room class")
room = create(api.parser.get_pobj_str("to"), parents=[room_class], location=None)

exits[direction] = {"door": door, "destination": room}

api.caller.location.set_property("exits", exits)
print(f'[color yellow]Created an exit to the {direction} to "{room.name}".[/color yellow]')



import pytest

from moo.core import code, parse
from moo.core.models import Object


@pytest.mark.django_db
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_basic_dig_and_tunnel(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.context(t_wizard, _writer):
        home_location = t_wizard.location
        parse.interpret("dig north to Another Room")
        assert printed == [
            '[color yellow]Created an exit to the north to "Another Room".[/color yellow]',
        ]
        printed.clear()

        parse.interpret("dig north to Another Room")
        assert printed == ["[color red]There is already an exit in that direction.[/color red]"]
        printed.clear()

        parse.interpret("go north")
        t_wizard.refresh_from_db()
        assert t_wizard.location.name == "Another Room"
        printed.clear()

        parse.interpret(f"tunnel south to {home_location.name}")
        assert printed == [
            f'[color yellow]Created an exit to the south to "{home_location.name}".[/color yellow]',
        ]
        printed.clear()

        parse.interpret("go south")
        t_wizard.refresh_from_db()
        assert t_wizard.location.name == home_location.name
        printed.clear()



# -*- coding: utf-8 -*-
"""
Core entity models for DjangoMOO
"""

from .acl import *
from .auth import *
from .object import AccessibleObject as Object
from .object import Alias, Relationship
from .property import AccessibleProperty as Property
from .task import *
from .verb import AccessibleVerb as Verb
from .verb import Repository, URLField, VerbName



import pytest

from moo.core import api, code, create, lookup, parse
from moo.core.models import Object


def setup_doors(t_wizard: Object):
    rooms = lookup("room class")
    room = create("Test Room", parents=[rooms])
    doors = lookup("door class")
    door = create("wooden door", parents=[doors], location=room)
    t_wizard.location = room
    t_wizard.save()
    api.caller.refresh_from_db()
    return room, door


@pytest.mark.django_db
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_creation(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.context(t_wizard, _writer):
        room, door = setup_doors(t_wizard)
        parse.interpret("dig north to Another Room through wooden door")
        assert printed == ['[color yellow]Created an exit to the north to "Another Room".[/color yellow]']
        assert t_wizard.location == room
        assert room.has_property("exits")
        assert room.exits["north"]["door"] == door

        printed.clear()
        parse.interpret("go north")
        api.caller.refresh_from_db()
        assert printed == ["You go north."]


@pytest.mark.django_db
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_locking(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.context(t_wizard, _writer):
        _, door = setup_doors(t_wizard)
        parse.interpret("dig north to Another Room through wooden door")
        assert printed == ['[color yellow]Created an exit to the north to "Another Room".[/color yellow]']
        printed.clear()
        parse.interpret("lock wooden door")
        assert printed == ["The door is locked."]
        assert door.is_locked()
        printed.clear()
        parse.interpret("unlock wooden door")
        assert printed == ["The door is unlocked."]
        assert not door.is_locked()


@pytest.mark.django_db
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_open(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.context(t_wizard, _writer):
        _, door = setup_doors(t_wizard)
        parse.interpret("dig north to Another Room through wooden door")
        assert printed == ['[color yellow]Created an exit to the north to "Another Room".[/color yellow]']
        printed.clear()
        parse.interpret("open wooden door")
        assert printed == ["The door is open."]
        assert door.is_open()
        printed.clear()
        parse.interpret("close wooden door")
        assert printed == ["The door is closed."]
        assert not door.is_open()



# -*- coding: utf-8 -*-
from .base import *  # pylint: disable=wildcard-import

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-061+p62f39ohlfrgu&)%1lxo%%#_-$rc5l_zsrlx6jqy)sw(=r"



#!moo verb is_open is_locked --on "door class" --method

from moo.core import api

door = api.parser.this

if args[0] == "is_open":
    prop_name = "open"
else:
    prop_name = "locked"

if door.has_property(prop_name) and door.get_property(prop_name):
    return True  # pylint: disable=return-outside-function  # type: ignore
else:
    return False  # pylint: disable=return-outside-function  # type: ignore



#!moo verb describe --on "System Object" --method

from moo.core import api

obj = args[1]  # pylint: disable=undefined-variable  # type: ignore

if obj.has_verb("describe", method=True):
    return obj.describe()  # pylint: disable=return-outside-function  # type: ignore
elif obj.has_property("description"):
    description = obj.get_property("description")
    return f"[bright_yellow]{obj.name}[/bright_yellow]\n[deep_sky_blue1]{description}[/deep_sky_blue1]"  # pylint: disable=return-outside-function  # type: ignore
else:
    return "[deep_pink4 bold]Not much to see here.[/deep_pink4 bold]"  # pylint: disable=return-outside-function  # type: ignore



# -*- coding: utf-8 -*-
"""
Verb model
"""

import logging

from django.core import validators
from django.db import models

from .. import utils
from ..code import interpret
from .acl import AccessibleMixin

log = logging.getLogger(__name__)


class Verb(models.Model, AccessibleMixin):
    #: The Python code for this Verb
    code = models.TextField(blank=True, null=True)
    #: Optional Git repo this code is from
    repo = models.ForeignKey("Repository", related_name="+", blank=True, null=True, on_delete=models.SET_NULL)
    #: Optional name of the code file within the repo
    filename = models.CharField(max_length=255, blank=True, null=True)
    #: Optional Git ref of the code file within the repo
    ref = models.CharField(max_length=255, blank=True, null=True)
    #: The owner of this Verb. Changes require `entrust` permission.
    owner = models.ForeignKey("Object", related_name="+", blank=True, null=True, on_delete=models.SET_NULL)
    #: The object on which this Verb is defined
    origin = models.ForeignKey("Object", related_name="verbs", on_delete=models.CASCADE)
    #: If True, this verb can only be used by the object it is defined on
    ability = models.BooleanField(default=False)
    #: If True, this verb can be invoked by other verbs
    method = models.BooleanField(default=False)

    def __str__(self):
        return "%s {#%s on %s}" % (self.annotated(), self.id, self.origin)

    @property
    def kind(self):
        return "verb"

    def annotated(self):
        ability_decoration = ["", "@"][int(self.ability)]
        method_decoration = ["", "()"][int(self.method)]
        verb_name = self.name()
        return "".join([ability_decoration, verb_name, method_decoration])

    def name(self):
        names = self.names.all()
        if not names:
            return "(untitled)"
        return names[0].name

    def save(self, *args, **kwargs):
        needs_default_permissions = self.pk is None
        super().save(*args, **kwargs)
        if not needs_default_permissions:
            return
        utils.apply_default_permissions(self)


class AccessibleVerb(Verb):
    class Meta:
        proxy = True

    def __call__(self, *args, **kwargs):
        if not (self.method):
            raise RuntimeError("%s is not a method." % self)
        if hasattr(self, "invoked_name"):
            l = list(args)
            l.insert(0, self.invoked_name)
            args = tuple(l)
        result = interpret(self.code, *args, **kwargs)
        return result


class VerbName(models.Model):
    verb = models.ForeignKey(Verb, related_name="names", on_delete=models.CASCADE)
    name = models.CharField(max_length=255, db_index=True)

    class Meta:
        constraints = [models.UniqueConstraint("verb", "name", name="unique_verb_name")]

    def __str__(self):
        return "%s {#%s on %s}" % (self.name, self.verb.id, self.verb.origin)


# TODO: add support for additional URL types and connection details
class URLField(models.CharField):
    default_validators = [validators.URLValidator(schemes=["https"])]


class Repository(models.Model):
    slug = models.SlugField()
    url = URLField(max_length=255)
    prefix = models.CharField(max_length=255)
